package uk.co.farowl.demo.reflection;

import java.lang.reflect.Method;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.util.LinkedList;
import java.util.List;

/**
 * This investigates http://bugs.jython.org/issue2662 in which reflective access to methods in a
 * class not exported to the application is prevented although the method implements a public API of
 * a class that is so exported. Run the program with JVM argument <code>--illegal-access=deny</code>
 * or <code>--illegal-access=warn</code> to demonstrate the problem.
 * <p>
 * Explore with reflective access when code in class <code>C</code> of package <code>P</code> holds
 * a reference to an object <code>x</code> that is an instance of a class <code>D</code> from an
 * inaccessible package <code>Q</code>. <code>D</code> extends or implements <code>E</code> from an
 * accessible package <code>R</code>. Suppose the code in <code>C</code> needs to invoke a method
 * <code>m</code> from <code>E</code>: we may obtain a <code>Method</code> for <code>D.m</code> by
 * reflection on <code>x</code>, but are prevented from invoking it because the defining class
 * <code>D</code> is from the inaccessible package <code>Q</code>. In a Java program, the compile
 * time type of x would be <code>E</code> or an accessible sub-type of <code>E</code>, and the
 * generated code would invoke <code>E.m</code>.
 * <p>
 * A couple of ideas are tried here that involve going up the inheritance chain (MRO) from the
 * actual class of the target object to find a method with the same signature that is amenable to a
 * reflective invocation at run-time.
 */
public class AccessAPI {

    public static void main(String[] args) {

        try {
            CharBuffer text = CharBuffer.wrap("hello");
            Charset gbk = Charset.forName("GBK");

            // This is perfectly legal
            naturalEncode(gbk, text);

            // This results in an "Illegal reflective access"
            try {
                Object enc = reflectiveNewEncoder(gbk);
                reflectiveUseEncoder(enc, text);
            } catch (IllegalAccessException iae) {
                iae.printStackTrace();
            }

            // How can I avoid that? Go to up the MRO to the first definition, assumed to be API.
            Object enc2 = apiNewEncoder(gbk);
            reflectiveUseEncoder(enc2, text);

            // How can I avoid that? Go up the MRO to the first definition we may set accessible.
            Object enc3 = accessibleNewEncoder(gbk);
            reflectiveUseEncoder(enc3, text);

        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    /** Get an encoder by reflection (first accessible version). */
    static Object accessibleNewEncoder(Charset cs)
            throws CharacterCodingException, ReflectiveOperationException {
        System.out.println("*** accessibleNewEncoder ***");

        // Reflectively get an encoder (from the highest-level public API)
        Method m = firstAccessible(cs.getClass(), "newEncoder", new Class<?>[0]);

        // Now work with the accessible method
        System.out.println("newEncoder method = " + m);
        Object enc = m.invoke(cs);

        return enc;
    }

    /** Find a method we'll be allowed to invoke on an instance of c. */
    static Method firstAccessible(Class<?> c, String name, Class<?>... args)
            throws NoSuchMethodException {
        /** Find the first accessible definition of the method in the MRO. */
        Method m = null;
        for (Class<?> ct : javaMRO(c)) {
            try {
                m = ct.getMethod(name, args);
                // Attempt to make accessible as way to check module access (Java 9 only)
                if (m.trySetAccessible()) {
                    break;
                }
            } catch (NoSuchMethodException | SecurityException e) {
                // Keep looking
            }
        }

        if (m == null) {
            throw new NoSuchMethodException(name);
        }
        return m;
    }

    /** Get an encoder by reflection (last in MRO version). */
    static Object apiNewEncoder(Charset cs)
            throws CharacterCodingException, ReflectiveOperationException {
        System.out.println("*** apiNewEncoder ***");

        // Reflectively get an encoder (from the highest-level public API)
        Method m = mostSuperior(cs.getClass(), "newEncoder", new Class<?>[0]);

        // Now work with the accessible method
        System.out.println("newEncoder method = " + m);
        Object enc = m.invoke(cs);

        return enc;
    }

    /** Find the most inherited method that matches, assuming it to be the public API. */
    static Method mostSuperior(Class<?> c, String name, Class<?>... args)
            throws NoSuchMethodException {

        /** Find the last matching definition of the method in the MRO. */
        Method m = null;
        for (Class<?> ct : javaMRO(c)) {
            try {
                m = ct.getMethod(name, args);
            } catch (NoSuchMethodException | SecurityException e) {
                // Keep looking
            }
        }

        if (m == null) {
            throw new NoSuchMethodException(name);
        }
        return m;
    }

    /** Generate the list of parent classes and interfaces. */
    private static List<Class<?>> javaMRO(Class<?> c) {
        LinkedList<Class<?>> mro = new LinkedList<>();
        for (Class<?> cls = c; cls != null;) {
            mro.add(cls);
            for (Class<?> iface : cls.getInterfaces()) {
                mro.add(iface);
            }
            cls = cls.getSuperclass();
        }
        return mro;
    }

    /** Get an encoder by reflection. */
    static Object reflectiveNewEncoder(Charset cs)
            throws CharacterCodingException, ReflectiveOperationException {
        System.out.println("*** reflectiveNewEncoder ***");

        // Reflectively get an encoder.
        Class<?> cc = cs.getClass();
        System.out.println("Charset class = " + cc);
        Method m = cc.getMethod("newEncoder", new Class<?>[0]);
        System.out.println("newEncoder method = " + m);
        // Illegal access here because cc is sun.nio.cs.GBK.
        Object enc = m.invoke(cs);

        return enc;
    }

    /** Use an encoder by reflection. */
    static void reflectiveUseEncoder(Object enc, CharBuffer text)
            throws CharacterCodingException, ReflectiveOperationException {
        text.rewind();
        Class<?> ec = enc.getClass();

        System.out.println("Encoder class = " + ec); // = sun.nio.cs.DoubleByte$Encoder
        Method em = ec.getMethod("encode", CharBuffer.class);
        System.out.println("Encode method = " + em);
        // No problem as em is java.nio.charset.CharsetEncoder.encode(CharBuffer)
        Object buf = em.invoke(enc, text);

        System.out.println("result = " + buf);
    }

    /** Get and use an encoder the natural way. */
    static void naturalEncode(Charset cs, CharBuffer text) throws CharacterCodingException {

        System.out.println("*** naturalEncode ***");
        text.rewind();

        CharsetEncoder enc = cs.newEncoder();
        ByteBuffer buf = enc.encode(text);

        System.out.println("result = " + buf);
    }

}
